/*
** User v1.2
**
** A WWW User CGI Enabler
**
** Copyright (c) 18.8.94 by Andreas Ley <ley@rz.uni-karlsruhe.de>
**
** Permission to use, copy, modify, and distribute this software for any
** purpose and without fee is hereby granted, provided that the above
** copyright notice appears in all copies. This software is provided "as is"
** and without any express or implied warranties.
**
** This program has been tested on a HP9000/715 with HP-UX A.09.05
** In this environment, neither lint -u nor gcc -Wall produce any messages.
** If you encounter any errors or need to make any changes to port it
** to another platform, please contact me.
**
** Usage:
**    - user must be setuid root to change uid
**    - If you don't need access control, just undefine ACCESS,	otherwise point
**	it to a configuration file (e.g. user.conf) with entries:
**		<username>
**	Any user with its name in the config file can now call scripts from his
**	.cgi-bin with a URL of the form http://host/cgi-bin/user/username/script
**	If you're using CERN httpd, you could create a httpd.conf entry reading:
**	Map	/~<star>/cgi-bin/<star>	/cgi-bin/user/<star>/<star>
**	and use http://host/~username/cgi-bin/script
**	(<star> is *, but I can't use * and / in a C comment :( )
**
** Version history
**
** Version 1.2 - 22.2.96
**	Added SERVER_NAME to SCRIPTNAME to allow for the alias mechanism to
**	be used with user CGIs.
**
** Version 1.1 - 16.11.95
**	Added support for unrestricted access
**
** Version 1.0 - 18.8.94
**	Initial version
*/

char header[]="user v1.2\n(c) 1994 by Andreas Ley\n";

/* General definitions */

#define	MASTER	"webmaster@rz.uni-karlsruhe.de"

/* File locations */

/*
#define ACCESS	"/usr/local/etc/httpd/config/user.conf"
*/
#define	USERDIR	".public_html"
#define	USERCGI	".cgi-bin"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <pwd.h>
#include <errno.h>
#include "user_german.h"


#define	FALSE	0		/* This is the naked Truth */
#define	TRUE	1		/* and this is the Light */

/* Global variables */

char	master[]=MASTER;


/* Prints header of a HTML-Request */
void show_header(char *subtitle, char *description)
{
	(void)printf("Content-type: text/html\n\n");
	(void)printf("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">\n");
	(void)printf("<HTML>\n");
	(void)printf("<HEAD>\n");
	(void)printf("<TITLE>%s%s%s</TITLE>\n",TITLE,subtitle?" - ":"",subtitle?subtitle:"");
	(void)printf("<LINK REV=\"made\" HREF=\"mailto:%s\">\n",master);
	(void)printf("</HEAD>\n");
	(void)printf("<BODY>\n");
	(void)printf("<H1>%s</H1>\n",subtitle?subtitle:TITLE);
	if (description) {
		(void)printf(description,master);
		(void)printf("<P>\n");
	}
}



/* Prints trailer of a HTML-Request */
void show_trailer()
{
	(void)printf("</BODY>\n");
	(void)printf("</HTML>\n");
}



/* Shows an error page, based on wrong user input */
void show_error(char *error)
{
	show_header(ERROR,error);
	show_trailer();
	exit(0);
}



/* Shows an error page, based on internal misfunction */
void show_fatal(char *error)
{
	show_header(ERROR,error);
	(void)printf(ERROR_FATAL,master);
	(void)printf("<P>\n");
	show_trailer();
	exit(0);
}



void usage(char *image)
{
	(void)fprintf(stderr,"Usage: %s [-h] [-v]\n",image);
	exit(1);
}


int main(int argc, char *argv[])
{
	int		c;
	char		scriptpath[MAXPATHLEN],buffer[8*BUFSIZ],*ptr,*username,*scriptname,*pathinfo;
	FILE		*src;
	struct passwd	*pwent;
#ifdef ACCESS
	int		found;
#endif

	while ((c=getopt(argc,argv,"vh?")) != EOF)
		switch ((char)c) {
		case 'v':
			(void)fprintf(stderr,header);
			exit(0);
		case 'h':
			(void)fprintf(stderr,header);
		case '?':
			usage(argv[0]);
		}

	if (!(username=getenv("PATH_INFO")))
		show_fatal(ERROR_USERNAME);
	if (*username=='/')
		username++;
	if (*username=='~')
		username++;
	if (!(scriptname=strchr(username,'/')))
		show_fatal(ERROR_SCRIPTNAME);
	*scriptname++='\0';
	if ((pathinfo=strchr(scriptname,'/')))
		*pathinfo++='\0';

#ifdef ACCESS
	if ((src=fopen(ACCESS,"r"))) {
		found=FALSE;
		while(fgets(buffer,sizeof(buffer),src)) {
			if (*buffer!='#') {
				if ((ptr=strchr(buffer,'\n')))
					*ptr='\0';
				if ((found=!strcmp(buffer,username)))
					break;
			}
		}
		fclose(src);
		if (!found) {
			(void)sprintf(buffer,ERROR_NOACCESS,username,master);
			show_error(buffer);
		}
	}
	else
		show_fatal(ERROR_READ);
#endif

	if (!(pwent=getpwnam(username)))
		show_fatal(ERROR_PASSWD);

	if (strncmp(getenv("SCRIPT_NAME"),"/~",2))
		(void)sprintf(buffer,"SCRIPT_NAME=%s/%s/%s",getenv("SCRIPT_NAME"),username,scriptname);
	else
		(void)sprintf(buffer,"SCRIPT_NAME=/~%s/cgi-bin/%s",username,scriptname);
	(void)putenv(strdup(buffer));

	if (pathinfo) {
		(void)sprintf(buffer,"PATH_INFO=/%s",pathinfo);
		(void)putenv(strdup(buffer));
		(void)sprintf(buffer,"PATH_TRANSLATED=%s/%s/%s",pwent->pw_dir,USERDIR,pathinfo);
		(void)putenv(strdup(buffer));
	}
	else {
		(void)putenv("PATH_INFO=");
		(void)putenv("PATH_TRANSLATED=");
	}

	if (setgid(pwent->pw_gid))
		show_fatal(ERROR_SETGID);
	if (setuid(pwent->pw_uid))
		show_fatal(ERROR_SETUID);

	(void)sprintf(scriptpath,"%s/%s/",pwent->pw_dir,USERCGI) ;
	if(chdir(scriptpath))
		show_fatal(ERROR_DIRECTORY);
	(void)strcat(scriptpath,scriptname) ;
	execv(scriptpath,argv);

	sprintf(buffer,ERROR_EXEC,scriptpath,strerror(errno));
	show_fatal(buffer);
	return(0);
}
